import sys
import select

def read_line() -> str | None:
    if sys.stdin in select.select([sys.stdin], [], [], 0)[0]:
        return sys.stdin.readline()

    return None


def write_line(line: str):
    sys.stdout.write(line)


from smulib import SMU, smu_script

@smu_script(name="Test")
def test(smu_instance: SMU, v: float, *args, **kwargs) -> dict[str, float]:
    smu_instance.set_voltage(v)

    return {
        "v": smu_instance.measure_voltage(),
        "a": smu_instance.measure_current()
    }


smu = SMU("/dev/cu.usbmodem11401")

res = smu.run_script(test, v=4.20)
print(res)


from machine import I2C, Pin

DAC_RANGE_16 = 65535

# ERROR CODES
DAC8571_OK                = 0x00
DAC8571_I2C_ERROR         = 0x81
DAC8571_ADDRESS_ERROR     = 0x82
DAC8571_BUFFER_ERROR      = 0x83

# WRITE MODES
DAC8571_MODE_STORE_CACHE  = 0x00
DAC8571_MODE_NORMAL       = 0x01
DAC8571_MODE_WRITE_CACHE  = 0x02
DAC8571_MODE_BRCAST_0     = 0x03
DAC8571_MODE_BRCAST_1     = 0x04   # not supported
DAC8571_MODE_BRCAST_2     = 0x05   # not supported

# DAC VALUES (percentages)
DAC8571_VALUE_00          = 0x0000
DAC8571_VALUE_25          = 0x4000
DAC8571_VALUE_50          = 0x8000
DAC8571_VALUE_75          = 0xC000
DAC8571_VALUE_100         = 0xFFFF

# POWER DOWN MODES
DAC8571_PD_LOW_POWER      = 0x00
DAC8571_PD_FAST           = 0x01
DAC8571_PD_1_KOHM         = 0x02
DAC8571_PD_100_KOHM       = 0x03
DAC8571_PD_HI_Z           = 0x04

def map_value(x, in_min, in_max, out_min, out_max):
    return (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min

class DAC8571:
    def __init__(self, i2c: I2C, address: int = 0x4C):
        self.__i2c = i2c
        self.__address = address
        self.__range = DAC_RANGE_16
        self.__dac = 0
        self.__control = 0
        self.__error = DAC8571_OK
        self.set_write_mode(DAC8571_MODE_NORMAL)

    def begin(self, value: int = DAC8571_VALUE_00) -> bool:
        if self.__address not in [0x4C, 0x4E]:
            self.__error = DAC8571_ADDRESS_ERROR
            return False

        if not self.is_connected():
            return False

        return self.write(value)

    def is_connected(self) -> bool:
        try:
            self.__i2c.writeto(self.__address, b"")
            return True

        except OSError:
            return False

    def get_address(self) -> int:
        return self.__address

    def write_range(self, value: float, r_min: float, r_max: float):
        if value < r_min:
            value = r_min

        if value > r_max:
            value = r_max

        self.write(map_value(value, r_min, r_max, 0, self.__range))

    def write(self, value: int) -> bool:
        low_byte = value & 0xFF
        high_byte = value >> 8

        buf = bytes([self.__control, high_byte, low_byte])

        try:
            self.__i2c.writeto(self.__address, buf)

        except OSError:
            self.__error = DAC8571_I2C_ERROR
            return False

        self.__error = DAC8571_OK
        self.__dac = value

        return True

    def write_mapped(self, v: float, r: tuple[float, float]) -> bool:
        v = int(map_value(v, r[0], r[1], 0, DAC_RANGE_16))

        return self.write(v)

    def read(self) -> int:
        data = self.__i2c.readfrom(self.__address, 3)

        high_byte = data[0]
        low_byte = data[1]

        self.__error = DAC8571_OK

        return high_byte * 256 + low_byte

    def set_percentage(self, percentage: float) -> bool:
        if percentage < 0:
            percentage = 0

        if percentage > 100:
            percentage = 100

        return self.write(int(percentage * 655.35))

    def get_percentage(self) -> float:
        return self.__dac * 0.0015259022

    def set_write_mode(self, mode: int = DAC8571_MODE_NORMAL) -> None:
        if mode > DAC8571_MODE_BRCAST_0:
            mode = DAC8571_MODE_NORMAL

        self.__control = mode << 4

    def get_write_mode(self) -> int:
        return (self.__control >> 4) & 0x03

    def power_down(self, pd_mode: int = DAC8571_PD_LOW_POWER) -> bool:
        pd_mask = {
            DAC8571_PD_LOW_POWER: 0x0000,
            DAC8571_PD_FAST: 0x2000,
            DAC8571_PD_1_KOHM: 0x4000,
            DAC8571_PD_100_KOHM: 0x8000,
            DAC8571_PD_HI_Z: 0xC000,
        }.get(pd_mode, 0x000)

        self.__control = (DAC8571_MODE_NORMAL << 4) + 0x01
        return self.write(pd_mask)

    def wake_up(self, value: int = DAC8571_VALUE_00) -> bool:
        self.set_write_mode(DAC8571_MODE_NORMAL)
        return self.write(value)

    def last_error(self) -> int:
        e = self.__error
        self.__error = DAC8571_OK
        return e

if __name__ == "__main__":
    i2c = I2C(0, scl=Pin(5), sda=Pin(4), freq=100_000)

    print([hex(d) for d in i2c.scan()])

    v_dac = DAC8571(i2c, 0x4E)

    t_v = 15 # Target voltage

    v_dac.write_mapped(t_v, (-15, 15))




from abc import ABC, abstractmethod

class SMUBase(ABC):
    @property
    @abstractmethod
    def port(self) -> str:
        pass

    @abstractmethod
    def measure_voltage(self) -> float:
        pass

    @abstractmethod
    def measure_current(self) -> float:
        pass

    def reconnect(self) -> None:
        raise NotImplementedError


import unittest
import random
import os
import time
from smulib import SMU
from smulib import SMULogger

class MockSMU(SMU):
    def __init__(self):
        self._port = 'MOCKPORT'

    @property
    def port(self):
        return self._port

    def measure_voltage(self):
        return random.uniform(-15, 15)

    def measure_current(self):
        return random.uniform(-15, 15)

    def reconnect(self):
        # do nothing for test
        pass

class TestSmuLoggerQueue(unittest.TestCase):
    def test_write_and_rotate_by_date(self):
        mock = MockSMU()
        td = os.curdir
        logger = SMULogger(smu=mock, data_request_interval=0.05, log_voltage=True, log_current=False,
                            data_output_path=td, batch_size=5, batch_timeout=0.1)
        try:
            logger.start_data_logging()
            time.sleep(0.5)
            logger.stop_data_logging()
            # wait a bit to let writer flush
            logger.close()
            # check files in data dir
            files = os.listdir(os.path.join(td, mock.port))
            self.assertTrue(len(files) >= 1)
            # find the csv file
            csv_files = [f for f in files if f.endswith('.csv')]
            self.assertTrue(len(csv_files) >= 1)
            fn = os.path.join(td, mock.port, csv_files[0])
            with open(fn, 'r', encoding='utf-8') as fh:
                lines = [l.strip() for l in fh.readlines() if l.strip()]
            # header + at least one row
            self.assertGreaterEqual(len(lines), 2)
            self.assertIn('timestamp_utc', lines[0])
        finally:
            # ensure stop cleanup
            logger.close()
    # end test

if __name__ == '__main__':
    unittest.main()



import logging
import time
import threading
from typing import Any, TYPE_CHECKING

import serial
from serial.tools import list_ports

from .base import SMUBase

if TYPE_CHECKING:
    from . import SMUScript, SMUScriptCallable

_LOG = logging.getLogger(__name__)

class SMU(SMUBase):
    def __init__(self, port: str, baudrate: int = 115200, timeout: float | None = 1.0):
        self.__port = port
        self.__baudrate = baudrate
        self.__timeout = timeout
        self._serial = None
        self._connect_attempts: int = 0
        self._last_connect_time: float = 0.0
        self.max_backoff: float = 3.0
        self.initial_backoff: float = 0.1

        self._serial_lock = threading.Lock()

        try:
            self.connect()
        except Exception:
            _LOG.exception("Initial connect failed for %s", port)
            self._serial = None

    @property
    def port(self) -> str:
        return self.__port

    def connect(self) -> None:
        with self._serial_lock:
            try:
                self._serial = serial.Serial(port=self.__port, baudrate=self.__baudrate, timeout=self.__timeout)
                self._connect_attempts = 0
                self._last_connect_time = time.monotonic()
                _LOG.debug("Connected to SMU on %s", self.__port)
            except Exception as e:
                # _LOG.exception("Failed to open serial port %s", self.__port)
                self._serial = None
                raise IOError(f"Could not open port {self.__port}: {e}") from e

    def disconnect(self) -> None:
        with self._serial_lock:
            if self._serial is not None:
                try:
                    self._serial.close()
                except Exception:
                    pass
                    # _LOG.exception("Error while closing serial port")
                finally:
                    self._serial = None
                    _LOG.debug("Disconnected from %s", self.__port)

    def reconnect(self) -> None:
        # with self._serial_lock: # removed because of deadlock with self.connect()
        backoff = self.initial_backoff
        attempts = 0
        while True:
            try:
                self.connect()
                return
            except IOError:
                attempts += 1
                if backoff > self.max_backoff:
                    raise IOError(f"Failed to reconnect after {attempts} attempts") from None
                # _LOG.warning("Reconnect attempt %d failed, backing off %.2fs", attempts, backoff)
                time.sleep(backoff)
                backoff *= 2

    def __write(self, command: str) -> None:
        # with self._serial_lock:
        if self._serial is None:
            raise IOError("Serial not connected")
        self._serial.write((command + "\n").encode())

    def __read(self, command: str) -> str:
        # with self._serial_lock:
        if self._serial is None:
            raise IOError("Serial not connected")
        self.__write(command)
        raw = self._serial.readline()
        if isinstance(raw, bytes):
            raw = raw.decode(errors='ignore')
        return str(raw).strip()

    def set_voltage(self, v: float) -> None:
        with self._serial_lock:
            self.__write(f":SOUR:VOLT {float(v)}")

    def set_current(self, i: float) -> None:
        with self._serial_lock:
            self.__write(f":SOUR:CURR {float(i)}")

    def voltage_limit(self) -> float:
        with self._serial_lock:
            return float(self.__read(":SOUR:VOLT:LIM?"))

    def current_limit(self) -> float:
        with self._serial_lock:
            return float(self.__read(":SOUR:CURR:LIM?"))

    def measure_voltage(self) -> float:
        with self._serial_lock:
            return float(self.__read(":MEAS:VOLT?"))

    def measure_current(self) -> float:
        with self._serial_lock:
            return float(self.__read(":MEAS:CURR?"))

    def run_script(self, script: SMUScriptCallable, *args, **kwargs) -> Any | list | dict:
        return script(self, *args, **kwargs)
    
    @staticmethod
    def discover_scripts(scripts_dir: str) -> list[SMUScript]:
        return []

    @staticmethod
    def discover_ports() -> dict[str, str]:
        """
        :return: A list of port-devices (e.g., ['COM3', '/dev/ttyUSB0'])
        """
        ports = list_ports.comports()
        return {port.device: port.description for port in ports}



import sys
import select
import time
import random

state = {
    "mode": "VOLT",
    "source_voltage": 0.0,
    "source_current": 0.0,
    "voltage_limit": 5.0,
    "current_limit": 0.1,
    "output": False,
}

def scpi_process(cmd: str) -> str:
    """Process a single SCPI command and return reply (if any)."""
    cmd = cmd.strip().upper()

    # Identification
    if cmd == "*IDN?":
        return "TU Delft,PicoSMU,42069,0.1.0\n"

    # Mode control
    elif cmd.startswith(":SOUR:FUNC?"):
        return state["mode"] + "\n"
    elif cmd.startswith(":SOUR:VOLT") and not cmd.endswith("?"):
        # Set voltage
        try:
            val = float(cmd.split()[-1])
            state["source_voltage"] = val
            state["mode"] = "VOLT"
            return ""
        except ValueError:
            return "ERR\n"
    elif cmd.startswith(":SOUR:CURR") and not cmd.endswith("?"):
        # Set current
        try:
            val = float(cmd.split()[-1])
            state["source_current"] = val
            state["mode"] = "CURR"
            return ""
        except ValueError:
            return "ERR\n"

    # Limits
    elif cmd.startswith(":SOUR:VOLT:LIM"):
        if cmd.endswith("?"):
            return f"{state['voltage_limit']:.4f}\n"
        else:
            try:
                state["voltage_limit"] = float(cmd.split()[-1])
                return ""
            except ValueError:
                return "ERR\n"
    elif cmd.startswith(":SOUR:CURR:LIM"):
        if cmd.endswith("?"):
            return f"{state['current_limit']:.6f}\n"
        else:
            try:
                state["current_limit"] = float(cmd.split()[-1])
                return ""
            except ValueError:
                return "ERR\n"

    # Measurement
    elif cmd == ":MEAS:VOLT?":
        # Return simulated voltage around source_voltage
        v = state["source_voltage"] + random.uniform(-0.1, 0.1)
        return f"{v:.4f}\n"
    elif cmd == ":MEAS:CURR?":
        # Return simulated current around source_current
        i = state["source_current"] + random.uniform(-0.1, 0.1)
        return f"{i:.6f}\n"

    # Output
    elif cmd.startswith(":OUTP"):
        if cmd.endswith("?"):
            return "1\n" if state["output"] else "0\n"
        else:
            val = cmd.split()[-1]
            state["output"] = val in ("1", "ON", "TRUE")
            return ""

    # Reset
    elif cmd == "*RST":
        state.update({
            "mode": "VOLT",
            "source_voltage": 0.0,
            "source_current": 0.0,
            "voltage_limit": 5.0,
            "current_limit": 0.1,
            "output": False,
        })
        return ""

    else:
        return "ERR:UNKNOWN\n"

def main():
    while True:
        if sys.stdin in select.select([sys.stdin], [], [], 0)[0]:
            line = sys.stdin.readline()
            if line:
                reply = scpi_process(line)
                if reply:
                    sys.stdout.write(reply)
        time.sleep(0.01)


if __name__ == "__main__":
    main()


import functools
from typing import Protocol, Any, runtime_checkable, Literal, TYPE_CHECKING

if TYPE_CHECKING:
    from . import SMU

def smu_script(name: str):
    def _decorator(func):
        @functools.wraps(func)
        def wrapper(smu, *args, **kwargs):
            return func(smu, *args, **kwargs)

        wrapper._is_smu_script = True
        wrapper._smu_script_name = name

        return wrapper

    return _decorator

@runtime_checkable
class SMUScriptCallable(Protocol):
    def __call__(self, smu: SMU, *args, **kwargs) -> Any | list | dict:
        ...

class SMUScriptParameter:
    def __init__(self, name: str, value_type: Literal["str", "int", "float"], default_value: Any):
        self.name = name
        self.value_type = value_type
        self.default = default_value

class SMUScript:
    def __init__(self, name: str, parameters: list[SMUScriptParameter], callable: SMUScriptCallable):
        self.name = name
        self.parameters = parameters
        self.callable = callable


from __future__ import annotations
import threading
import logging
import time
import os
import csv
from datetime import datetime, timezone, date
from typing import Dict, Any, Optional
from logging.handlers import RotatingFileHandler

# Import locally to avoid circular imports if package structure allows, 
# otherwise assume smu class is available
# from ..smu import SMU 

CSV_HEADERS = ["timestamp_utc", "timestamp_epoch_ms", "voltage_v", "current_a"]

class SMULogger:
    def __init__(self,
                 smu, # Type: SMU | None
                 data_request_interval: float = 1.0,
                 log_voltage: bool = True,
                 log_current: bool = True,
                 data_dir: str | None = None,
                 batch_size: int = 100,
                 batch_timeout: float = 0.5,
                 debug_log_dir: str | None = None,
                 debug_max_file_size_bytes: int = 5 * 1024 * 1024,
                 debug_backup_count: int = 5,
                 encoding: str = 'utf-8'):

        self.smu = smu
        self.data_request_interval = data_request_interval
        self.log_voltage = log_voltage
        self.log_current = log_current
        self.encoding = encoding
        
        # Determine initial safe port name
        if self.smu:
             self.safe_port = str(getattr(self.smu, 'port', 'unknown')).replace('/', '_').replace('\\', '_')
        else:
             self.safe_port = "unknown"

        if data_dir is None: self.data_dir = os.path.join(os.path.abspath('.'), "data")
        else: self.data_dir = os.path.abspath(data_dir)
        os.makedirs(self.data_dir, exist_ok=True)

        if not debug_log_dir: debug_log_dir = os.path.join(os.path.abspath('.'), "logs")
        self._debug_log_path = os.path.join(debug_log_dir, 'smu_debug.log')
        os.makedirs(debug_log_dir, exist_ok=True)

        handler = RotatingFileHandler(
            self._debug_log_path,
            maxBytes=debug_max_file_size_bytes,
            backupCount=debug_backup_count,
            encoding=self.encoding
        )
        formatter = logging.Formatter('%(asctime)s - %(levelname)s\t- %(message)s')
        handler.setFormatter(formatter)

        self.debug_logger = logging.getLogger(f'smu.debug')
        self.debug_logger.propagate = False
        # Clean up existing handlers
        if self.debug_logger.hasHandlers():
            self.debug_logger.handlers.clear()
            
        self.debug_logger.addHandler(handler)
        self.debug_logger.setLevel(logging.DEBUG)

        # threads and control
        self._poll_thread = threading.Thread(target=self._poll_loop, daemon=True)
        self._stop_event = threading.Event()
        self._running_event = threading.Event()
        self._settings_lock = threading.Lock()

    def start_data_logging(self) -> bool:
        if self.smu is None:
            self.debug_logger.debug('No SMU selected, cannot start logging')
            return False
        
        self._stop_event.clear()
        
        if not self._poll_thread.is_alive():
            # Create new thread if old one died or hasn't started
            self._poll_thread = threading.Thread(target=self._poll_loop, daemon=True)
            self._poll_thread.start()
            
        self._running_event.set()
        self.debug_logger.info('Data logging started')
        return True

    def stop_data_logging(self):
        self.debug_logger.info('Stopping data logging...')
        self._running_event.clear() # Pause logic
        self.debug_logger.info('Data logging paused/stopped')

    def shutdown(self):
        """Completely stop threads for application exit"""
        self._running_event.clear()
        self._stop_event.set()
        try:
            self._poll_thread.join(timeout=1)
        except Exception:
            pass

    def set_logging_parameters(self, smu = None, log_voltage: Optional[bool] = None, log_current: Optional[bool] = None, data_request_interval: Optional[float] = None):
        with self._settings_lock:
            if smu is not None:
                self.smu = smu
                # Update safe port for filename
                self.safe_port = str(getattr(self.smu, 'port', 'unknown')).replace('/', '_').replace('\\', '_')
            if log_voltage is not None:
                self.log_voltage = log_voltage
            if log_current is not None:
                self.log_current = log_current
            if data_request_interval is not None:
                self.data_request_interval = data_request_interval
        
        self.debug_logger.debug('Parameters updated. SMU: %s', "Attached" if self.smu else "None")

    def _poll_loop(self):
        file_header = None
        csv_writer = None
        current_date = None

        def close_current_file():
            nonlocal file_header
            if file_header:
                try:
                    file_header.flush()
                    file_header.close()
                except Exception:
                    pass
                file_header = None

        def open_todays_file():
            nonlocal current_date, file_header, csv_writer
            close_current_file()

            current_date = datetime.now(timezone.utc).date()
            self._ensure_header_for_date(current_date)
            file_name = self._current_filename_for_date(current_date)
            try:
                file_header = open(file_name, 'a', encoding=self.encoding, newline='')
                csv_writer = csv.DictWriter(file_header, fieldnames=CSV_HEADERS)
            except Exception as e:
                self.debug_logger.error(f"Failed to open log file {file_name}: {e}")
                file_header = None

        self.debug_logger.debug('Poller thread running')
        open_todays_file()

        while not self._stop_event.is_set():
            loop_start = time.monotonic()
            
            # 1. Check if we should be logging
            if not self._running_event.is_set():
                time.sleep(0.2)
                continue
                
            # 2. Check date rotation
            if datetime.now(timezone.utc).date() != current_date:
                open_todays_file()

            # 3. Snapshot settings
            with self._settings_lock:
                current_smu = self.smu
                log_v = self.log_voltage
                log_c = self.log_current
                interval = self.data_request_interval

            if current_smu is None:
                time.sleep(0.5)
                continue

            # 4. Measure
            now = datetime.now(timezone.utc)
            row: Dict[str, Any] = {
                CSV_HEADERS[0]: now.isoformat(),
                CSV_HEADERS[1]: int(now.timestamp() * 1000),
                CSV_HEADERS[2]: '',
                CSV_HEADERS[3]: ''
            }
            
            valid_read = False
            try:
                # We check connected status first to avoid timeouts if known dead
                # but SMU methods also throw IOErrors if dead.
                if log_v:
                    row['voltage_v'] = current_smu.measure_voltage()
                if log_c:
                    row['current_a'] = current_smu.measure_current()
                valid_read = True
            except Exception:
                # If reading fails (disconnect), we just skip writing this row.
                # We do NOT reconnect here; the Main UI thread handles reconnection logic.
                # We just wait to be told to resume or for the SMU object to be swapped.
                pass

            # 5. Write if valid
            if valid_read and file_header and csv_writer:
                try:
                    csv_writer.writerow(row)
                    file_header.flush()
                except Exception:
                    pass

            # 6. Sleep remaining time
            elapsed = time.monotonic() - loop_start
            to_sleep = max(0.0, interval - elapsed)
            if self._stop_event.wait(to_sleep):
                break

        close_current_file()
        self.debug_logger.debug('Poller thread exiting')

    def _current_filename_for_date(self, d: date) -> str:
        # Uses self.safe_port which is updated in set_logging_parameters
        return os.path.join(self.data_dir, f"{self.safe_port}-{d.isoformat()}.csv")

    def _ensure_header_for_date(self, d: date):
        filename = self._current_filename_for_date(d)
        try:
            os.makedirs(os.path.dirname(filename), exist_ok=True)
            if os.path.exists(filename) and os.path.getsize(filename) != 0: return
            with open(filename, 'a', encoding=self.encoding, newline='') as fh:
                writer = csv.DictWriter(fh, fieldnames=CSV_HEADERS)
                writer.writeheader()
        except Exception:
            self.debug_logger.exception('Failed to ensure header for %s', filename)


import time

from machine import I2C, Pin

from .lib import DAC8571
from .serial import read_line, write_line

i2c = I2C(0, scl=Pin(5), sda=Pin(4), freq=400_000)

voltage_dac = DAC8571(i2c, 0x4C)
current_dac = DAC8571(i2c, 0x4C)

def process_line(line: str) -> str:
    pass

def main():
    while True:
        line = read_line()
        reply = process_line(line)
        write_line(reply)

        time.sleep(0.01)


if __name__ == "__main__":
    main()